module App

    class AWSCloudFormation

        # Gets the path to the root of the cloudformation templates in blufin-secrets repo.
        # @return string
        def self.get_cloudformation_path
            if App::AWSProfile::get_profile[App::AWSProfile::CLOUDFORMATION]['Templates'].has_key?('Local')
                path = App::AWSProfile::get_profile['CloudFormation']['Templates']['Local']['Path']
                Blufin::Terminal::error("CloudFormation path not found: #{Blufin::Terminal::format_directory(path)}", 'This probably means directories got renamed or code got changed.', true) unless Blufin::Files::path_exists(path)
            else
                path = download_s3_templates
            end
            path
        end

        # Uploads a cloudformation template to S3 (so we can create a stack from it).
        # Returns the S3 URL.
        # @return string
        def self.upload_cloudformation_template(source, category, template, filename_override: nil)
            raise RuntimeError, "File does not exist: #{source}" unless Blufin::Files::file_exists(source)
            tmp_file = "/tmp/aws-cf-upload-#{category}-#{template}-#{Blufin::Strings::random_string}.yml"
            Blufin::Terminal::execute("cp #{source} #{tmp_file}", text: "Preparing template: \x1B[38;5;240m#{tmp_file}\x1B[0m")
            bucket_path = get_s3_bucket_path
            bucket_path = bucket_path == '' ? '' : "#{bucket_path}/"
            if filename_override.nil?
                template_filename = "#{category}-#{template}-#{DateTime.now.strftime('%Y%m%d-%H%M%S')}.yml"
            else
                template_filename = filename_override
            end
            App::AWSCli::s3_upload(tmp_file, App::AWSProfile::get_profile['CloudFormation']['Uploads']['S3Bucket']['Name'], "#{bucket_path}#{template_filename}")
            # Return the S3 bucket URL.
            "#{get_s3_bucket_url}/#{template_filename}"
        end

        # Returns S3 bucket URL based on config parameters.
        # @return string
        def self.get_s3_bucket_url
            bucket_path = get_s3_bucket_path
            bucket_path = bucket_path == '' ? '' : "/#{bucket_path}"
            "https://#{App::AWSProfile::get_profile['CloudFormation']['Uploads']['S3Bucket']['Name']}.s3-#{App::AWSProfile::get_profile['CloudFormation']['Uploads']['S3Bucket']['Region']}.amazonaws.com#{bucket_path}"
        end

        # Returns the S3 bucket path based on config parameters.
        # @return string
        def self.get_s3_bucket_path
            bucket_path = App::AWSProfile::get_profile['CloudFormation']['Uploads']['S3Bucket']['Path']
            bucket_path.nil? || bucket_path.to_s.strip == '' ? '' : bucket_path
        end

        # Returns an array of cloudformation stack-names.
        # @return Array of Hashes
        def self.get_stacks
            results = []
            json    = Blufin::Terminal::execute('awx l -r CloudFormationStacks -j', capture: true, text: "AWS \xe2\x80\x94 Getting CloudFormation Stacks...")
            begin
                data = JSON.parse(json)
            rescue => e
                puts
                puts json
                Blufin::Terminal::error('Failed to parse AWS response JSON.', e.message)
            end
            data.each do |stack|
                results << {
                    :name   => stack['StackName'],
                    :region => stack['region']
                }
            end
            results.sort_by! { |hsh| hsh[:name] }
            results
        end

        # Downloads templates from S3 into /tmp folder (and returns path to /tmp folder).
        # @return string
        def self.download_s3_templates(use_cache = true)
            raise RuntimeError, "Missing key in #{App::CONFIG_FILE}: #{App::AWSProfile::CLOUDFORMATION}->Templates->S3Bucket" unless App::AWSProfile::get_profile[App::AWSProfile::CLOUDFORMATION]['Templates'].has_key?('S3Bucket')
            s3       = App::AWSProfile::get_profile[App::AWSProfile::CLOUDFORMATION]['Templates']['S3Bucket']
            tmp_path = Blufin::AWS::download_s3_data(s3['Name'], s3['Path'], profile: App::AWSProfile::get_profile[App::AWSProfile::PROFILE], region: s3['Region'], use_cache: use_cache)
            tmp_path
        end

    end

end