require 'date'
require 'columnist'

module App

    class AWSOutputter

        EC2_INSTANCES       = 'ec2_instances'
        EC2_VOLUMES         = 'ec2_volumes'
        EC2_SECURITY_GROUPS = 'ec2_security_groups'
        S3_BUCKETS          = 's3_buckets'
        RDS_MYSQL_INSTANCES = 'rds_mysql_instances'
        SQS_QUEUES          = 'sqs_queues'

        HEADER_COLOR     = 255
        DIVIDER_COLOR    = 240
        DIM_COLOR        = 246
        CONSTRAINT_COLOR = 67

        REGION_WIDTH = 15

        extend Columnist

        # Outputs EC2 data to terminal.
        # @return void
        def self.display(resource_title, resource, column_data, output, total_width, opts, args)

            output = [] if output.nil?

            @terminal_width = Blufin::Terminal::get_terminal_width

            wildcard_width = @terminal_width - total_width - 8 - (column_data.length)

            puts "  \x1B[38;5;255m\x1B[48;5;54m #{output.length} \x1B[48;5;18m #{resource_title} \x1B[0m"
            puts
            table(:border => false) do

                row do
                    column('', :width => 2, :color => HEADER_COLOR)
                    column_data.each do |title, data|
                        column(title, :width => get_width(data[:width], wildcard_width), :color => HEADER_COLOR)
                    end
                    column('', :width => 2, :color => HEADER_COLOR)
                end

                row do
                    column('')
                    column_data.each do |title, data|
                        column('-' * get_width(data[:width], wildcard_width), :color => DIVIDER_COLOR)
                    end
                    column('')
                end

                output.each do |item|
                    row do
                        column('')
                        column_data.each do |title, data|
                            color = data[:color].nil? ? 'default' : data[:color]
                            if data[:key] =~ /^Tags\./
                                value = App::AWSReports::extract_tag(item, data[:key])
                            else
                                value = get_value(data[:key], item)
                            end
                            value, color = get_formatter(data[:formatter], resource: resource, color: color).call(value) unless data[:formatter].nil?
                            column(value, :color => get_color(color))
                        end
                        column('')
                    end
                end

            end
            puts

        end

        # Displays CF parameters in a table (currently during the CF intro screen).
        # @return void
        def self.display_parameters(table_data)
            raise RuntimeError, "Expected Array, instead got: #{table_data.class}" unless table_data.is_a?(Array)
            return unless table_data.any?
            # Must match widths below (minus remaining).
            tw            = Blufin::Terminal::get_terminal_width
            column_short  = 5
            column_widths = [32, 15, 75, 25, column_short, column_short, column_short, column_short]
            remaining     = tw - column_widths.inject(0) { |sum, x| sum + x } - 16
            column_widths << remaining
            table(:border => false) do
                row do
                    column('', :width => 2, :color => HEADER_COLOR)
                    column('Parameter', :width => 32, :color => HEADER_COLOR)
                    column('Type', :width => 15, :color => HEADER_COLOR)
                    column('Description', :width => 75, :color => HEADER_COLOR)
                    column('AllowedPattern', :width => 25, :color => HEADER_COLOR)
                    column('Min-L', :width => column_short, :color => HEADER_COLOR)
                    column('Max-L', :width => column_short, :color => HEADER_COLOR)
                    column('Min-V', :width => column_short, :color => HEADER_COLOR)
                    column('Max-V', :width => column_short, :color => HEADER_COLOR)
                    column('Cached/Default Value(s)', :width => remaining, :color => HEADER_COLOR)
                    column('', :width => 2, :color => HEADER_COLOR)
                end
                row do
                    column('')
                    column_widths.each do |width|
                        column('-' * width, :color => DIVIDER_COLOR)
                    end
                    column('')
                end
                table_data.each do |data|
                    alp = data[:allowed_pattern]
                    mil = data[:min_length]
                    miv = data[:min_value]
                    mal = data[:max_length]
                    mav = data[:max_value]
                    cop = 'light-green'
                    cop = 'purple' if data[:type_internal] == :autocomplete
                    cop = 'light-grey' if data[:type_internal] == :system || [
                        AppCommand::CloudFormationCreate::OPTION_STACK_NAME,
                        AppCommand::CloudFormationCreate::OPTION_PROJECT_ID,
                        AppCommand::CloudFormationCreate::OPTION_DESCRIPTION,
                        AppCommand::CloudFormationCreate::OPTION_TERM_PROTECT,
                        AppCommand::CloudFormationCreate::OPTION_TIMEOUT,
                        AppCommand::CloudFormationCreate::OPTION_OWNER
                    ].include?(data[:parameter])
                    # In case you want to change color in future, currently this is redundant.
                    cod = data[:type_internal] == :system ? 'default' : 'default'
                    cot = data[:type_internal] == :system ? 'default' : 'default'
                    cot = data[:type] == AppCommand::CloudFormationCreate::SPECIAL ? 'light-grey' : cot
                    row do
                        column('')
                        column(data[:parameter], :color => get_color(cop))
                        column(data[:type], :color => get_color(cot))
                        column(data[:description], :color => get_color(cod))
                        column(alp.nil? ? "\xe2\x80\x94" : alp, :color => alp.nil? ? get_color('default') : CONSTRAINT_COLOR)
                        column(mil.nil? ? "\xe2\x80\x94".rjust(column_short, ' ') : mil.to_s.rjust(column_short, ' '), :color => mil.nil? ? get_color('default') : CONSTRAINT_COLOR)
                        column(mal.nil? ? "\xe2\x80\x94".rjust(column_short, ' ') : mal.to_s.rjust(column_short, ' '), :color => mal.nil? ? get_color('default') : CONSTRAINT_COLOR)
                        column(miv.nil? ? "\xe2\x80\x94".rjust(column_short, ' ') : miv.to_s.rjust(column_short, ' '), :color => miv.nil? ? get_color('default') : CONSTRAINT_COLOR)
                        column(mav.nil? ? "\xe2\x80\x94".rjust(column_short, ' ') : mav.to_s.rjust(column_short, ' '), :color => mav.nil? ? get_color('default') : CONSTRAINT_COLOR)
                        column(data[:cached_value], :color => get_color('green'))
                        column('')
                    end
                end
            end
        end

        private

        # Recursively get nested keys. Does not handle Arrays.
        # @return string
        def self.get_value(key, item)
            return '—' if item.nil? || !item.is_a?(Hash)
            ks = key.split('.')
            fk = ks[0]
            ks.shift
            if ks.length == 0
                return '—' unless item.has_key?(key) # If key not found, return mdash.
                item[fk]
            else
                get_value(ks.join('.'), item[fk])
            end
        end

        # Define all custom formatters here.
        # @return void
        def self.get_formatter(key = nil, resource: nil, color: 'default')
            formatters = {
                'cloudformation-stack'    => Proc.new { |n|
                    [n.gsub(/^arn:aws:cloudformation:[a-z]{2}-[a-z]{2,12}-\d+\D?:\d{10,14}:stack\//, ''), color]
                },
                'cloudfront-status'       => Proc.new { |n|
                    case n
                        when 'Deployed'
                            color = 'green'
                        when 'InProgress'
                            color = 'orange'
                        else
                            Blufin::Terminal::output("Add something to aws_outputter switch statement for: #{n}", Blufin::Terminal::MSG_TODO)
                            color = 'default'
                    end
                    [n, color]
                },
                'ec2-security-groups'     => Proc.new { |n|
                    security_groups = []
                    if n.any?
                        n.each do |security_group|
                            group_name = security_group['GroupName']
                            group_name = 'N/A' if group_name.strip == '' || group_name.nil?
                            security_groups << group_name
                        end
                    end
                    security_groups.sort!
                    security_group = security_groups.include?('default') ? 'default' : security_groups[0]
                    security_group = "#{security_group}+#{security_groups.length - 1}" if security_groups.length > 1
                    [security_group, 'default']
                },
                'ec2-status'              => Proc.new { |n|
                    case n
                        when 'pending'
                            color = 'orange'
                        when 'running'
                            color = 'green'
                        when 'in-use'
                            color = 'green'
                        when 'stopped'
                            color = 'dark-grey'
                        when 'terminated'
                            color = 'red'

                        else
                            Blufin::Terminal::output("Add something to aws_outputter switch statement for: #{n}", Blufin::Terminal::MSG_TODO)
                            color = 'default'
                    end
                    [n, color]
                },
                'epoch-date'              => Proc.new { |n|
                    [DateTime.strptime(n.to_i.to_s, '%s'), color]
                },
                'region'                  => Proc.new { |n|
                    preferred_regions = resource.nil? ? [] : resource[App::AWSReports::KEY_REGIONS_PREFERRED]
                    region_color      = preferred_regions.include?(n) ? 'brown' : 'red'
                    region            = n.ljust(10, ' ')
                    [region, region_color]
                },
                'route-53-domain-name'    => Proc.new { |n|
                    [n.gsub(/\.$/, ''), color]
                },
                'route-53-hosted-zone-id' => Proc.new { |n|
                    zone_color = color
                    ns         = n.split('/')
                    [ns[ns.length - 1], zone_color]
                }
            }
            raise RuntimeError, "Key not found: #{key}" unless key.nil? || formatters.has_key?(key)
            return formatters.keys if key.nil?
            formatters[key]
        end

        # Define all custom colors here.
        # @return void
        def self.get_color(key = nil)
            colors = {
                'default'     => 240,
                'red'         => 196,
                'orange'      => 202,
                'yellow'      => 220,
                'green'       => 46,
                'blue'        => 39,
                'purple'      => 61,
                'pink'        => 198,
                'brown'       => 94,
                'light-grey'  => 246,
                'light-green' => 77,
                'dark-grey'   => 240,
            }
            raise RuntimeError, "Key not found: #{key}" unless key.nil? || colors.has_key?(key)
            return colors.keys if key.nil?
            colors[key]
        end

        # Determines whether to get width or wildcard width.
        # @return
        def self.get_width(width, wildcard_width)
            (width == '*') ? wildcard_width : width
        end

        # Standardized way of outputting a command.
        # @return void
        def self.output_cli_command(command, path = nil)
            puts "  \x1B[38;5;70m$ \x1B[38;5;240m#{command}\x1B[0m" if path.nil?
            puts "  \x1B[38;5;70m$ \x1B[38;5;240m#{command} [\x1B[38;5;246m#{path}\x1B[240m]\x1B[0m" unless path.nil?
        end

        # Standardized way of rendering selection output.
        # @return string
        def self.render_selection(primary, secondary = nil, tertiary = nil)
            return "#{Blufin::Terminal::format_action(primary)} \x1B[38;5;240m[#{Blufin::Terminal::format_highlight(secondary)}\x1B[38;5;240m]\x1B[38;5;240m[#{Blufin::Terminal::format_highlight(tertiary)}\x1B[38;5;240m]" if !tertiary.nil? && !secondary.nil?
            return "#{Blufin::Terminal::format_action(primary)} \x1B[38;5;240m[#{Blufin::Terminal::format_highlight(secondary)}\x1B[38;5;240m]" unless secondary.nil?
            return "#{Blufin::Terminal::format_action(primary)}" if secondary.nil?
        end

    end

end